//******************************************
// Ewan Ward
// Copyright (C) Heavy Horse Research
// Program: OwnerDraw.exe
// FILE: MainDlg.CPP
//
// NOTES:
// The functions PrepareImageRect() and DrawTheIcon()
// are taken from the CButtonST class by Davide Calabro
//
// Bug Fixes and enhancements
// 1. Double-click messages posted as single clicks - suggestion by John Wellbelove
// 2. When using Classic style in XP owner draw button not drawn - bug found by  Zhefu Zhang
// 3. Does not respond to change of theme.
//******************************************

#include "stdafx.h"
#include "resource.h"

// XP specific stuff
#include <uxtheme.h>
#include <tmschema.h>



extern HINSTANCE ghInst;
extern HWND ghMainWnd;
extern HWND ghDlgCurrent;
static HICON hIcon = NULL;
static HICON hOwnerDrawIcon = NULL;


#define ICON_HEIGHT 38
#define ICON_WIDTH 38

// XP Themes
BOOL Themed = FALSE;
BOOL ThemeLibLoaded = FALSE;

typedef HRESULT(__stdcall *PFNCLOSETHEMEDATA)(HTHEME hTheme);
typedef HRESULT(__stdcall *PFNDRAWTHEMEBACKGROUND)(HTHEME hTheme, HDC hdc, 
							int iPartId, int iStateId, const RECT *pRect,  const RECT *pClipRect);
typedef HTHEME(__stdcall *PFNOPENTHEMEDATA)(HWND hwnd, LPCWSTR pszClassList);
typedef HRESULT (__stdcall *PFNDRAWTHEMETEXT)(HTHEME hTheme, HDC hdc, int iPartId, 
							int iStateId, LPCWSTR pszText, int iCharCount, DWORD dwTextFlags, 
							DWORD dwTextFlags2, const RECT *pRect);
typedef HRESULT (__stdcall *PFNGETTHEMEBACKGROUNDCONTENTRECT)(HTHEME hTheme,  HDC hdc, 
							int iPartId, int iStateId,  const RECT *pBoundingRect, 
							RECT *pContentRect);


PFNOPENTHEMEDATA zOpenThemeData = NULL;
PFNDRAWTHEMEBACKGROUND zDrawThemeBackground = NULL;
PFNCLOSETHEMEDATA zCloseThemeData = NULL;
PFNDRAWTHEMETEXT zDrawThemeText = NULL;
PFNGETTHEMEBACKGROUNDCONTENTRECT zGetThemeBackgroundContentRect = NULL;

static HMODULE hModThemes = NULL;

void InitThemes()
{
	hModThemes = LoadLibrary("UXTHEME.DLL");
	if(hModThemes)
	{
		zOpenThemeData = (PFNOPENTHEMEDATA)GetProcAddress(hModThemes, "OpenThemeData");
		zDrawThemeBackground = (PFNDRAWTHEMEBACKGROUND)GetProcAddress(hModThemes, "DrawThemeBackground");
		zCloseThemeData = (PFNCLOSETHEMEDATA)GetProcAddress(hModThemes, "CloseThemeData");
		zDrawThemeText = (PFNDRAWTHEMETEXT)GetProcAddress(hModThemes, "DrawThemeText");
		zGetThemeBackgroundContentRect = (PFNGETTHEMEBACKGROUNDCONTENTRECT)GetProcAddress(hModThemes, "GetThemeBackgroundContentRect");

		if(zOpenThemeData 
			&& zDrawThemeBackground
			&& zCloseThemeData
			&& zDrawThemeText
			&& zGetThemeBackgroundContentRect)
		{
			ThemeLibLoaded = TRUE;			
		}
		else
		{
			FreeLibrary(hModThemes);
			hModThemes = NULL;
		}
	}
}

void FinThemes()
{
	FreeLibrary(hModThemes);
	hModThemes = NULL;
	ThemeLibLoaded = FALSE;
	Themed = FALSE;
}




// helpers

static void PrepareImageRect(HWND hButtonWnd, BOOL bHasTitle, RECT* rpItem, RECT* rpTitle, BOOL bIsPressed, DWORD dwWidth, DWORD dwHeight, RECT* rpImage)
{
	RECT rBtn;

	CopyRect(rpImage, rpItem);


	GetClientRect(hButtonWnd, &rBtn);
	if (bHasTitle == FALSE)
	{
		// Center image horizontally
		LONG rpImageWidth = rpImage->right - rpImage->left;
		rpImage->left += ((rpImageWidth - (long)dwWidth)/2);
	}
	else
	{
		// Image must be placed just inside the focus rect
		LONG rpTitleWidth = rpTitle->right - rpTitle->left;
		rpTitle->right = rpTitleWidth - dwWidth - 30;
		rpTitle->left = 30;
		rpImage->left = rBtn.right - dwWidth - 30;
		// Center image vertically
		LONG rpImageHeight = rpImage->bottom - rpImage->top;
		rpImage->top += ((rpImageHeight - (long)dwHeight)/2);
	}
		
	// If button is pressed then press image also
	if (bIsPressed && !Themed)
		OffsetRect(rpImage, 1, 1);
} // End of PrepareImageRect

static void DrawTheIcon(HWND hButtonWnd, HDC* dc, BOOL bHasTitle, RECT* rpItem, RECT* rpTitle, BOOL bIsPressed, BOOL bIsDisabled)
{
	RECT	rImage;
	PrepareImageRect(hButtonWnd, bHasTitle, rpItem, rpTitle, bIsPressed, ICON_WIDTH, ICON_HEIGHT, &rImage);

	// Ole'!
	DrawState(	*dc,
				NULL,
				NULL,
				(LPARAM)hOwnerDrawIcon,
				0,
				rImage.left,
				rImage.top,
				(rImage.right - rImage.left),
				(rImage.bottom - rImage.top), 
				(bIsDisabled ? DSS_DISABLED : DSS_NORMAL) | DST_ICON);
} // End of DrawTheIcon


static LONG prev_proc;
// Message handler for wizard button
static BOOL bMouseOverButton = FALSE;

LRESULT CALLBACK ButtWindProc(
	HWND hWnd,                            /* window handle                   */
	UINT message,                         /* type of message                 */
	WPARAM wParam,                        /* additional information          */
	LPARAM lParam)                        /* additional information          */
	{
	switch (message)
		{
		case WM_LBUTTONDBLCLK:
			PostMessage(hWnd, WM_LBUTTONDOWN, wParam, lParam);
			break;
		case WM_MOUSEMOVE:
			{
			// check to see if the left mousr button is pressed
			if(!bMouseOverButton)
				{
				bMouseOverButton = TRUE;
				InvalidateRect(hWnd, NULL, FALSE);
				}

			return (0);
			}
			break;
		}
	// Any messages we don't process must be passed onto the original window function
	return CallWindowProc((WNDPROC)prev_proc, hWnd, message, wParam, lParam);

	}

// Mesage handler for main box.
LRESULT CALLBACK MainDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	static HTHEME hTheme;

	switch (message)
	{
		case WM_THEMECHANGED:
			{
			if(ThemeLibLoaded)
				{
				if(hTheme)
					zCloseThemeData(hTheme);
				FinThemes();					
				}

			InitThemes();
			if(ThemeLibLoaded)
				{
				hTheme = zOpenThemeData(GetDlgItem(hDlg, IDC_OWNERDRAW_BTN), L"Button");
				if(hTheme)
					Themed = TRUE;
				}
			}
			break;
		case WM_INITDIALOG:
			ghDlgCurrent = hDlg; // This is the only dialog!

			InitThemes();
			if(hIcon == NULL)
				hIcon = LoadIcon(ghInst, MAKEINTRESOURCE(IDI_OWNERDRAW));

			if(hOwnerDrawIcon == NULL)
				hOwnerDrawIcon = (HICON)LoadImage(ghInst, 
										MAKEINTRESOURCE(IDI_BUTTON_ICO), 
										IMAGE_ICON, 
										ICON_WIDTH,
										ICON_HEIGHT,
										0);

			// subclass the Wizard button so we can know when the mouse has moved over it
			prev_proc = SetWindowLongPtr(GetDlgItem(hDlg, IDC_OWNERDRAW_BTN), GWLP_WNDPROC, (LONG)ButtWindProc);

			if(ThemeLibLoaded)
				{
				hTheme = zOpenThemeData(GetDlgItem(hDlg, IDC_OWNERDRAW_BTN), L"Button");
				if(hTheme)
					Themed = TRUE;
				}

			SendMessage(hDlg, WM_SETICON, ICON_SMALL, (LPARAM)hIcon);

			SetFocus(GetDlgItem(hDlg, IDC_NORMAL_BTN));
			break;
		
		case WM_MOUSEMOVE:
			{
			if(bMouseOverButton)
				{
				bMouseOverButton = FALSE;
				InvalidateRect(GetDlgItem(hDlg, IDC_OWNERDRAW_BTN), NULL, TRUE);
				}
			}
			break;
		case WM_DRAWITEM:
			{
			// draw the owner draw button
			LPDRAWITEMSTRUCT lpDIS = (LPDRAWITEMSTRUCT) lParam;
			
			if(lpDIS->CtlID != IDC_OWNERDRAW_BTN)
				return (0);

			HDC dc = lpDIS->hDC;

			// button state
			BOOL bIsPressed =	(lpDIS->itemState & ODS_SELECTED);
			BOOL bIsFocused  = (lpDIS->itemState & ODS_FOCUS);
			BOOL bIsDisabled = (lpDIS->itemState & ODS_DISABLED);
			BOOL bDrawFocusRect = !(lpDIS->itemState & ODS_NOFOCUSRECT);

			RECT itemRect = lpDIS->rcItem;

			SetBkMode(dc, TRANSPARENT);

			// Prepare draw... paint button background

			if(Themed)
			{
				DWORD state = (bIsPressed)?PBS_PRESSED:PBS_NORMAL;
				
				if(state == PBS_NORMAL)
					{
					if(bIsFocused)
						state = PBS_DEFAULTED;
					if(bMouseOverButton)
						state = PBS_HOT;
					}
				zDrawThemeBackground(hTheme, dc, BP_PUSHBUTTON, state, &itemRect, NULL);
			}
			else
			{
				if (bIsFocused)
					{
					HBRUSH br = CreateSolidBrush(RGB(0,0,0));  
					FrameRect(dc, &itemRect, br);
					InflateRect(&itemRect, -1, -1);
					DeleteObject(br);
					} // if		
		
				COLORREF crColor = GetSysColor(COLOR_BTNFACE);

				HBRUSH	brBackground = CreateSolidBrush(crColor);

				FillRect(dc, &itemRect, brBackground);

				DeleteObject(brBackground);

				// Draw pressed button
				if (bIsPressed)
				{
					HBRUSH brBtnShadow = CreateSolidBrush(GetSysColor(COLOR_BTNSHADOW));
					FrameRect(dc, &itemRect, brBtnShadow);
					DeleteObject(brBtnShadow);
				}
		
				else // ...else draw non pressed button
				{
				    UINT uState = DFCS_BUTTONPUSH |
                          ((bMouseOverButton) ? DFCS_HOT : 0) |
                          ((bIsPressed) ? DFCS_PUSHED : 0);

					DrawFrameControl(dc, &itemRect, DFC_BUTTON, uState);
				} // else
			}
			// Read the button's title
			char sTitle[100];
			GetWindowText(GetDlgItem(hDlg, IDC_OWNERDRAW_BTN), sTitle, 100);

			RECT captionRect = lpDIS->rcItem;

			// Draw the icon
			BOOL bHasTitle = (sTitle[0] != '\0');
			DrawTheIcon(GetDlgItem(hDlg, IDC_OWNERDRAW_BTN), &dc, bHasTitle, &lpDIS->rcItem, &captionRect, bIsPressed, bIsDisabled);

			// Write the button title (if any)
			if (bHasTitle)
			{// Draw the button's title
				// If button is pressed then "press" title also
				if (bIsPressed && !Themed)
					OffsetRect(&captionRect, 1, 1);

				// Center text
				RECT centerRect = captionRect;
				DrawText(dc, sTitle, -1, &captionRect, DT_WORDBREAK | DT_CENTER | DT_CALCRECT);
				LONG captionRectWidth = captionRect.right - captionRect.left;
				LONG captionRectHeight = captionRect.bottom - captionRect.top;
				LONG centerRectWidth = centerRect.right - centerRect.left;
				LONG centerRectHeight = centerRect.bottom - centerRect.top;
				OffsetRect(&captionRect, (centerRectWidth - captionRectWidth)/2, (centerRectHeight - captionRectHeight)/2);

				if(Themed)
				{					
					// convert title to UNICODE obviously you don't need to do this if you are a UNICODE app.
					int nTextLen = strlen(sTitle); 
					int mlen = MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, (char *)sTitle, nTextLen + 1, NULL, 0); 
					WCHAR* output = new WCHAR[mlen];
					if(output)
					{
						MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, (char *)sTitle, nTextLen + 1, output, mlen);
						zDrawThemeText(	hTheme, dc, BP_PUSHBUTTON, PBS_NORMAL,
										output, wcslen(output),
										DT_CENTER | DT_VCENTER | DT_SINGLELINE,
										0, &captionRect);
						delete output;
					}
				}
				else
				{
					SetBkMode(dc, TRANSPARENT);

					if (bIsDisabled)
						{
						OffsetRect(&captionRect, 1, 1);
						SetTextColor(dc, ::GetSysColor(COLOR_3DHILIGHT));
						DrawText(dc, sTitle, -1, &captionRect, DT_WORDBREAK | DT_CENTER);
						OffsetRect(&captionRect, -1, -1);
						SetTextColor(dc, ::GetSysColor(COLOR_3DSHADOW));
						DrawText(dc, sTitle, -1, &captionRect, DT_WORDBREAK | DT_CENTER);
						} // if
					else
						{
						SetTextColor(dc, ::GetSysColor(COLOR_BTNTEXT));
						SetBkColor(dc, ::GetSysColor(COLOR_BTNFACE));
						DrawText(dc, sTitle, -1, &captionRect, DT_WORDBREAK | DT_CENTER);
						} // if
					} // if
			}

			// Draw the focus rect
			if (bIsFocused && bDrawFocusRect)
				{
				RECT focusRect = itemRect;
				InflateRect(&focusRect, -3, -3);
				DrawFocusRect(dc, &focusRect);
				} // if
			return (TRUE);
			}
			break;
		case WM_SETICON:
			DefWindowProc(hDlg, message, wParam, lParam);
			break;
		case WM_SYSCOMMAND:
			{
			return DefWindowProc(hDlg, message, wParam, lParam);
			}
		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
			case IDCANCEL:
			case IDOK:
				DestroyIcon(hIcon);
				DestroyIcon(hOwnerDrawIcon);
				if(ThemeLibLoaded)
				{
					if(hTheme)
						zCloseThemeData(hTheme);
					FinThemes();
				}
				PostQuitMessage(0);
				return TRUE;
			}//switch
			break;
	}
    return FALSE;
}

HWND DoMainDlg(HWND parent)
{
	DWORD dwErr;
	HWND hRet = CreateDialog(ghInst, (LPCTSTR)IDD_MAIN_DLG, parent, (DLGPROC)MainDlgProc);
	if(hRet == NULL)
		dwErr = GetLastError();

	return hRet;

}

